#include "common.h"
#include <cstdlib>
#include <cstdio>

#include "tbb/parallel_for.h"
#include "tbb/task_scheduler_init.h"
#include "tbb/tick_count.h"

const size_t N = 3000000;
static std::string Data[N];

static const std::string adjective[] = {
    "sour",
    "sweet",
    "bitter",
    "salty",
    "big",
    "medium",
    "small"
};

const size_t n_adjective = sizeof(adjective)/sizeof(adjective[0]);

static const std::string noun[] = {
    "apple",
    "banana",
    "cherry",
    "date",
    "eggplant",
    "fig",
    "grape",
    "honeydew",
    "icao",
    "jujube",
    "kiwi",
    "lemon",
    "melon",
    "nectarine",
    "orange",
    "pear",
    "quince"
};

const size_t n_noun = sizeof(noun)/sizeof(noun[0]);

static void initialize() {
    for( int i=0; i<N; ++i ) {
        Data[i] = adjective[rand()%n_adjective];
        Data[i] += " ";
        Data[i] += noun[rand()%n_noun];
    }
}

void time_serial( const char* name, void (*clear)(), void (*update)( const std::string& key ) ) {
    for( int i=0; i<2; ++i ) {
        clear();
        tbb::tick_count before = tbb::tick_count::now();
        for( std::string* j=Data; j!=Data+N; ++j )
            SerialUpdate(*j);
        tbb::tick_count after = tbb::tick_count::now();
        if( i==1 ) 
            std::printf( "\t%s:\t%g seconds\n", name, (after-before).seconds ());
    }
}

void check( const char* name, int (*routine)( const std::string& key ) ) {
    for( int i=0; i<n_adjective; ++i ) {
        for( int j=0; j<n_noun; ++j ) {
            std::string key = adjective[i] + " " + noun[i];
            int x = SerialGet(key);
            int y = routine(key);
            if( x!=y ) {
                std::printf("Error for key '%s'\n"
                            "serial version computed %d occurrences\n"
                            "parallel version computed %d occurrences\n", 
                            key.c_str(), x, y );
                std::abort();
            }
        }
    }
}

void time_parallel( const char* name, void (*clear)(), void (*update)( const std::string& key ), int (*get)( const std::string& key ) ) {
    for( int i=0; i<2; ++i ) {
        clear();
	tbb::tick_count before = tbb::tick_count::now();
        try {
	    tbb::parallel_for(
		tbb::blocked_range<std::string*> (Data, Data+N), 
		[&]( const tbb::blocked_range<std::string*>& range ) {
		    for( std::string* i=range.begin(); i!=range.end(); ++i )
			update(*i);
		},
		tbb::auto_partitioner() );
        } catch(...) {
            std::printf("\t%s:\t(threw exception)\n", name );
            return;
        }
	tbb::tick_count after = tbb::tick_count::now();
        check( name, get );
        if( i==1 ) 
            std::printf( "\t%s:\t%g seconds\n", name, (after-before).seconds ());
    }
}

int main (void) {
    srand (2);
    initialize();

    time_serial( "Serial", SerialClear, SerialUpdate );
    for( int p=1; p<=4; ++p ) {
        tbb::task_scheduler_init init(p);
        printf("Running with %d threads\n", p);
        time_parallel( "CoarseGrain", CoarseGrainClear, CoarseGrainUpdate, CoarseGrainGet );
        time_parallel( "FineGrain", FineGrainClear, FineGrainUpdate, FineGrainGet );
    }
    return 0;
}
