#include "common.h"
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include "tbb/tick_count.h"

void Normalize( Digit z[], size_t n ) {
    assert(n>0);
    int carry = 0;
    for( size_t i=0; i<n-1; ++i ) {
        int total = z[i] + carry;
        if( total>=0 ) {
            carry = total/Radix;
        } else {
            carry = -1-int((-1-total)/Radix);
        }
        z[i] = total-carry*Radix;
        assert( 0<=z[i] );
        assert( z[i]<Radix );
    }
    z[n-1] += carry;
    assert( z[n-1]>=0 );
}

void SimpleMul( Digit z[], const Digit x[], const Digit y[], size_t n ) {
    for( size_t i=0; i<2*n; ++i )
        z[i] = 0;
    for( size_t i=0; i<n; ++i ) {
        for( size_t j=0; j<n; ++j ) {
            int t = z[i+j]+x[i]*y[j];
            assert(t==Digit(t));
            z[i+j] = Digit(t);
        }
    }
    Normalize(z,2*n);
}

static Digit A[MaxN], B[MaxN], C_Serial[2*MaxN], C_Parallel[2*MaxN];

static void initialize() {
    for (int i=0; i<MaxN; ++i) {
        for (int j=0; j<MaxN; ++j) {
            A[i] = 9;
            B[i] = 9;
        }
    }
}

static void check() {
    for (int i=0; i<2*MaxN; ++i) 
        if( C_Parallel[i]!=C_Serial[i] ) {
            std::printf("Error: C_Parallel[%d]=%d C_Serial[%d]=%d\n", i, C_Parallel[i], i, C_Serial[i] );
            std::abort();
        }
}

void time_one( const char* name, void (*routine)(Digit [], const Digit [], const Digit [], size_t ), Digit c[] ) {
    // Run once to warm up cache
    routine( c, A, B, MaxN );
    // Run again to time it.
    tbb::tick_count before = tbb::tick_count::now ();
    routine( c, A, B, MaxN );
    Normalize( c, 2*MaxN );
    tbb::tick_count after = tbb::tick_count::now ();
    std::printf( "%s:\t%g seconds\n", name, (after-before).seconds());
}

int main (void) {
    initialize();

    time_one( "SerialKaratsubaMul", SerialKaratsubaMul, C_Serial );
    time_one( "ParallelKaratsubaMul", ParallelKaratsubaMul, C_Parallel );

    check(); 
    return 0;
}
